import bisect
import calendar
import datetime
import logging
import math

from django.conf import settings

logger = logging.getLogger("django_calendardate")


fy_start_month = getattr(settings, "FISCAL_YEAR_START_MONTH", 10)


def process_date(date):
    """
    Returns a dictionary with all of the below processes completed,
    which can then be imported directly to CalendarDate model.
    """

    metadata = {}
    metadata["calendar_date"] = date
    metadata["calendar_day"] = get_calendar_day(date)
    metadata["calendar_month"] = get_calendar_month(date)
    metadata["calendar_year"] = get_calendar_year(date)
    metadata["calendar_quarter"] = get_calendar_quarter(date)
    metadata["fiscal_year"] = get_fiscal_year(date)
    metadata["fiscal_quarter"] = get_fiscal_quarter(date)
    metadata["day_of_week"] = get_day_of_week(date)
    metadata["day_of_isoweek"] = get_day_of_isoweek(date)
    metadata["day_of_quarter"] = get_day_of_quarter(date)
    metadata["day_of_year"] = get_day_of_year(date)
    metadata["week_of_month"] = get_week_of_month(date)
    metadata["week_of_year"] = get_week_of_year(date)
    metadata["isoweek_of_year"] = get_isoweek_of_year(date)
    metadata["is_weekday"] = get_is_weekday(date)
    metadata["is_leap_year"] = get_is_leap_year(date)
    metadata["days_in_month"] = get_days_in_month(date)

    logger.debug(f"process_date metadata: {metadata}")

    return metadata


def get_quarter_begin(date):
    quarter_begins = [datetime.date(date.year, month, 1) for month in (1, 4, 7, 10)]
    idx = bisect.bisect(quarter_begins, date)
    return quarter_begins[idx - 1]


def get_calendar_day(date):
    return date.day


def get_calendar_month(date):
    return date.month


def get_calendar_year(date):
    return date.year


def get_calendar_quarter(date):
    return int(math.ceil(date.month / 3))


def get_fiscal_year(date):
    return date.year if date.month < fy_start_month or fy_start_month == 1 else date.year + 1


def get_fiscal_quarter(date):

    if fy_start_month in [1, 4, 7, 10]:
        quarter_starts = [1, 4, 7, 10]
        quarter_idx = quarter_starts.index(fy_start_month)
    elif fy_start_month in [2, 5, 8, 11]:
        quarter_starts = [2, 5, 8, 11]
        quarter_idx = quarter_starts.index(fy_start_month)
    else:
        quarter_starts = [3, 6, 9, 12]
        quarter_idx = quarter_starts.index(fy_start_month)

    quarter_start_dates = [datetime.date(2020, month, 1) for month in quarter_starts]

    quarter = bisect.bisect(quarter_start_dates, date) - quarter_idx
    return quarter if quarter > 0 else quarter + 4


def get_day_of_week(date):
    return date.weekday()


def get_day_of_isoweek(date):
    return date.isoweekday()


def get_day_of_quarter(date):
    delta = date - get_quarter_begin(date) + datetime.timedelta(days=1)
    return delta.days


def get_day_of_year(date):
    return int(date.strftime("%j"))


def get_week_of_month(date):
    def week_of_month(date_object):
        """Returns the week of the month for the specified date."""
        first_day = date_object.replace(day=1)
        dom = date_object.day
        adjusted_dom = dom + first_day.weekday()
        return int(math.ceil(adjusted_dom / 7.0))

    return week_of_month(date)


def get_week_of_year(date):
    return int(date.strftime("%U")) + 1


def get_isoweek_of_year(date):
    return int(date.strftime("%W")) + 1


def get_is_weekday(date):
    return True if date.weekday() < 5 else False


def get_is_leap_year(date):
    year = date.year
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def get_days_in_month(date):
    _, days = calendar.monthrange(date.year, date.month)
    return days



"""
Based on the date range provided, this command adds CalendarDate
objects for any non-yet-existing dates in that range.
"""
import logging

from datetime import datetime, timedelta
from django.core.management.base import BaseCommand, CommandError
from ...models import CalendarDate
from ...date_utilities import process_date

logger = logging.getLogger("django_calendardate")


class Command(BaseCommand):
    help = """Processes (and creates) dates for the CalendarDate model.
        Use: `manage.py process_dates 2000-01-01 2005-12-31`

        Optionally, the --force argument will overwrite existing dates.
        """

    def add_arguments(self, parser):
        parser.add_argument("start_date", type=str, help="Start of the date period, formatted as: YYYY-MM-DD")
        parser.add_argument("end_date", type=str, help="End of the date period, formatted as: YYYY-MM-DD")

        parser.add_argument(
            "--force",
            action="store_true",
            help="Force overwrite of existing dates",
        )

    def handle(self, *args, **options):
        start_date = datetime.strptime(options["start_date"], "%Y-%m-%d").date()
        end_date = datetime.strptime(options["end_date"], "%Y-%m-%d").date()
        delta = timedelta(days=1)

        logger.debug(f"Processing dates from {start_date} to {end_date}")

        while start_date <= end_date:

            date_metadata = process_date(start_date)

            obj, created = CalendarDate.objects.update_or_create(**date_metadata)

            start_date += delta

            if created:
                self.stdout.write(
                    self.style.SUCCESS(f"Successfully created/verified CalendarDate for {obj.calendar_date}.")
                )
            else:
                if options["force"]:
                    self.stdout.write(self.style.SUCCESS(f"CalendarDate for {obj.calendar_date} updated."))
                else:
                    self.stdout.write(
                        self.style.SUCCESS(f"CalendarDate for {obj.calendar_date} already exists. Skipping.")
                    )



from django.contrib.gis.db import models
from django.utils.translation import gettext as _


class CalendarDate(models.Model):
    calendar_date = models.DateField(_("Date"), unique=True, help_text=_("The actual date object"))
    calendar_day = models.PositiveSmallIntegerField(_("Calendar Day"), help_text=_("Number from 1 through 31"))
    calendar_month = models.PositiveSmallIntegerField(_("Calendar Month"), help_text=_("Month number from 1-12"))
    calendar_year = models.PositiveSmallIntegerField(
        _("Calendar Year"), help_text=_("Current year, eg: 2017, 2025, 1984")
    )
    calendar_quarter = models.PositiveSmallIntegerField(
        _("Calendar Quarter"), help_text=_("1-4, indicates quarter within the current year")
    )
    fiscal_year = models.PositiveSmallIntegerField(
        _("Fiscal Year"), help_text=_("Current fiscal year, eg: 2017, 2025, 1984")
    )
    fiscal_quarter = models.PositiveSmallIntegerField(
        _("Fiscal Quarter"), help_text=_("1-4, indicates fiscal quarter within the current fiscal year")
    )
    day_of_week = models.PositiveSmallIntegerField(_("Day of Week"), help_text=_("Monday is 0 and Sunday is 6"))
    day_of_isoweek = models.PositiveSmallIntegerField(_("Day of ISO Week"), help_text=_("Monday is 1 and Sunday is 7"))
    day_of_quarter = models.PositiveSmallIntegerField(
        _("Day in Quarter"), help_text=_("Number from 1-92, indicates the day # in the current quarter")
    )
    day_of_year = models.PositiveSmallIntegerField(_("Day in Year"), help_text=_("Number from 1-366"))
    week_of_month = models.PositiveSmallIntegerField(
        _("Week of Month"), help_text=_("Number from 1-6, indicates the number of week within the current month")
    )
    week_of_year = models.PositiveSmallIntegerField(
        _("Week of Year"), help_text=_("Number from 1-53, indicates the number of week within the current year")
    )
    isoweek_of_year = models.PositiveSmallIntegerField(
        _("ISO Week of Year"), help_text=_("Number from 1-53, indicates the number of isoweek within the current year")
    )
    is_weekday = models.BooleanField(
        _("Is Weekday"), default=False, help_text=_("True if Monday-->Friday, False for Saturday/Sunday")
    )
    is_leap_year = models.BooleanField(
        _("Is Leap Year"), default=False, help_text=_("True if current year is a leap year")
    )
    days_in_month = models.PositiveSmallIntegerField(
        _("Days in current month"), help_text=_("Number of days in the current month")
    )

    class Meta:
        verbose_name = _("Calendar Date")
        verbose_name_plural = _("Calendar Dates")

        get_latest_by = ["-calendar_date"]
        ordering = ["-calendar_date"]

        indexes = [
            models.Index(fields=["calendar_date"]),
            models.Index(fields=["calendar_year", "calendar_month"]),
            models.Index(fields=["calendar_year", "calendar_quarter"]),
            models.Index(fields=["calendar_year", "week_of_year"]),
            models.Index(fields=["fiscal_year", "fiscal_quarter"]),
            models.Index(fields=["is_weekday"]),
        ]

    def __str__(self):
        return f"{self.date}"

    @property
    def get_day_name(self):
        day_names = {
            1: _("Monday"),
            2: _("Tuesday"),
            3: _("Wednesday"),
            4: _("Thursday"),
            5: _("Friday"),
            6: _("Saturday"),
            7: _("Sunday"),
        }
        return day_names[self.day_of_isoweek]

    @property
    def get_month_name(self):
        month_names = {
            1: _("January"),
            2: _("February"),
            3: _("March"),
            4: _("April"),
            5: _("May"),
            6: _("June"),
            7: _("July"),
            8: _("August"),
            9: _("September"),
            10: _("October"),
            11: _("November"),
            12: _("December"),
        }
        return month_names[self.calendar_month]

    @property
    def get_month_abbreviated(self):
        month_names = {
            1: _("Jan"),
            2: _("Feb"),
            3: _("Mar"),
            4: _("Apr"),
            5: _("May"),
            6: _("Jun"),
            7: _("Jul"),
            8: _("Aug"),
            9: _("Sep"),
            10: _("Oct"),
            11: _("Nov"),
            12: _("Dec"),
        }
        return month_names[self.calendar_month]



from django.apps import apps
from django.contrib import admin


# Autoregister model
class ListAdminMixin(object):
    def __init__(self, model, admin_site):
        self.list_display = [field.name for field in model._meta.fields]
        super().__init__(model, admin_site)


for model in apps.get_app_config("django_calendardate").get_models():
    admin_class = type("AdminClass", (ListAdminMixin, admin.ModelAdmin), {})
    try:
        admin.site.register(model, admin_class)
    except admin.sites.AlreadyRegistered:
        pass



import os
from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), "README.rst")) as fh:
    readme = fh.read()

setup(
    name="django-calendardate",
    version="0.1.2",
    description="A calendar model with date metadata for querying against.",
    long_description=readme,
    author="Jack Linke",
    author_email="jack@watervize.com",
    url="http://github.com/OmenApps/django-calendardate/",
    packages=find_packages(),
    package_data={
        "django_calendardate": [],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Office/Business :: Scheduling",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
    ],
)



# Generated by Django 3.1.7 on 2021-02-21 17:10

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CalendarDate",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "calendar_date",
                    models.DateField(help_text="The actual date object", unique=True, verbose_name="Date"),
                ),
                (
                    "calendar_day",
                    models.PositiveSmallIntegerField(help_text="Number from 1 through 31", verbose_name="Calendar Day"),
                ),
                (
                    "calendar_month",
                    models.PositiveSmallIntegerField(help_text="Month number from 1-12", verbose_name="Calendar Month"),
                ),
                (
                    "calendar_year",
                    models.PositiveSmallIntegerField(
                        help_text="Current year, eg: 2017, 2025, 1984", verbose_name="Calendar Year"
                    ),
                ),
                (
                    "calendar_quarter",
                    models.PositiveSmallIntegerField(
                        help_text="1-4, indicates quarter within the current year", verbose_name="Calendar Quarter"
                    ),
                ),
                (
                    "fiscal_year",
                    models.PositiveSmallIntegerField(
                        help_text="Current fiscal year, eg: 2017, 2025, 1984", verbose_name="Fiscal Year"
                    ),
                ),
                (
                    "fiscal_quarter",
                    models.PositiveSmallIntegerField(
                        help_text="1-4, indicates fiscal quarter within the current fiscal year",
                        verbose_name="Fiscal Quarter",
                    ),
                ),
                (
                    "day_of_week",
                    models.PositiveSmallIntegerField(
                        help_text="Monday is 0 and Sunday is 6", verbose_name="Day of Week"
                    ),
                ),
                (
                    "day_of_isoweek",
                    models.PositiveSmallIntegerField(
                        help_text="Monday is 1 and Sunday is 7", verbose_name="Day of ISO Week"
                    ),
                ),
                (
                    "day_of_quarter",
                    models.PositiveSmallIntegerField(
                        help_text="Number from 1-92, indicates the day # in the current quarter",
                        verbose_name="Day in Quarter",
                    ),
                ),
                (
                    "day_of_year",
                    models.PositiveSmallIntegerField(help_text="Number from 1-366", verbose_name="Day in Year"),
                ),
                (
                    "week_of_month",
                    models.PositiveSmallIntegerField(
                        help_text="Number from 1-6, indicates the number of week within the current month",
                        verbose_name="Week of Month",
                    ),
                ),
                (
                    "week_of_year",
                    models.PositiveSmallIntegerField(
                        help_text="Number from 1-53, indicates the number of week within the current year",
                        verbose_name="Week of Year",
                    ),
                ),
                (
                    "isoweek_of_year",
                    models.PositiveSmallIntegerField(
                        help_text="Number from 1-53, indicates the number of isoweek within the current year",
                        verbose_name="ISO Week of Year",
                    ),
                ),
                (
                    "is_weekday",
                    models.BooleanField(
                        default=False,
                        help_text="True if Monday-->Friday, False for Saturday/Sunday",
                        verbose_name="Is Weekday",
                    ),
                ),
                (
                    "is_leap_year",
                    models.BooleanField(
                        default=False, help_text="True if current year is a leap year", verbose_name="Is Leap Year"
                    ),
                ),
                (
                    "days_in_month",
                    models.PositiveSmallIntegerField(
                        help_text="Number of days in the current month", verbose_name="Days in current month"
                    ),
                ),
            ],
            options={
                "verbose_name": "Calendar Date",
                "verbose_name_plural": "Calendar Dates",
                "ordering": ["-calendar_date"],
                "get_latest_by": ["-calendar_date"],
            },
        ),
        migrations.AddIndex(
            model_name="calendardate",
            index=models.Index(fields=["calendar_date"], name="django_cale_calenda_95378f_idx"),
        ),
        migrations.AddIndex(
            model_name="calendardate",
            index=models.Index(fields=["calendar_year", "calendar_month"], name="django_cale_calenda_fcdc17_idx"),
        ),
        migrations.AddIndex(
            model_name="calendardate",
            index=models.Index(fields=["calendar_year", "calendar_quarter"], name="django_cale_calenda_b87451_idx"),
        ),
        migrations.AddIndex(
            model_name="calendardate",
            index=models.Index(fields=["calendar_year", "week_of_year"], name="django_cale_calenda_a55a54_idx"),
        ),
        migrations.AddIndex(
            model_name="calendardate",
            index=models.Index(fields=["fiscal_year", "fiscal_quarter"], name="django_cale_fiscal__592cc1_idx"),
        ),
        migrations.AddIndex(
            model_name="calendardate",
            index=models.Index(fields=["is_weekday"], name="django_cale_is_week_c191dd_idx"),
        ),
    ]
